#!/usr/bin/env python3
#
# Classes to access lava via the API
#

import requests
from ruamel.yaml import YAML
import re
import logging
import json

from lava_credentials import get_lava_host, get_lava_credentials


class LavaException(Exception):
    pass


class Lava:
    """Access functions for the lava rest api, the class
    holds the token and and the host.
    All other requests should be stateless.
    """

    def __init__(self, host, token):
        self._host = host
        self._token = token

    def plain_request(self, url):
        response = requests.get(
            url=url,
            headers={"Authorization": "Token %s" % self._token},
        )

        # Handle errors
        if not response.ok:
            logging.error("Response from lava %d", response.status_code)
            return response.ok, None
        logging.debug("Response from lava %d", response.status_code)

        return response.ok, response.text

    def request(self, suburl):
        # Get test results via LAVA REST API
        ok, text = self.plain_request(
            "http://{host}/api/v0.2/{suburl}".format(host=self._host, suburl=suburl)
        )
        return ok, text

    def request_json(self, suburl):
        ok, text = self.request(suburl)
        if not ok:
            raise LavaException("Failed to query lava suburl: %s" % suburl)

        results = json.loads(text)
        return results

    def get_job_suites(self, jobid):
        return self.request_json("jobs/{}/suites/".format(jobid))

    def get_job_details(self, jobid):
        return self.request_json("jobs/{}/".format(jobid))

    def get_test_results(self, jobid):
        return self.request_json("jobs/{}/tests/".format(jobid))

    def get_test_log(self, jobid):
        ok, content = self.request("jobs/{}/logs/".format(jobid))
        if not ok:
            raise LavaException("Failed to query lava for test log of job %d" % jobid)

        # print(content)
        yaml = YAML(typ="safe")
        log = yaml.load(content)

        return log

    def get_device_list(self):
        return self.request_json("devices/")

    def get_device(self, name):
        return self.request_json("devices/{}/".format(name))

    def get_devicetypes(self, name):
        return self.request_json("devicetypes/{}".format(name))

    def get_tag_list(self):
        return self.request_json("/tags/")


class LavaTest:
    """Abstraction for one lava test"""

    def __init__(self, parent, test, joblog):

        self._test = test
        self.parent = parent
        self.heading = None
        self.test_case_id = test["name"]
        self.result = self._test["result"] == "pass"
        self.measurement = self._test["measurement"]
        self.timestamp = self._test["logged"]

        log_end = test["end_log_line"]
        if log_end is None:
            logging.debug(test)
            raise LavaException("No logline defined, skipping this test")

        (
            log_start,
            log_end,
            description_start,
            description_end,
        ) = self._search_log_for_test(joblog, log_end)

        self.log = joblog[log_start:log_end]
        self.description = ""

        if description_start is not None and description_end is not None:
            for line in joblog[description_start + 1 : description_end]:
                if line["lvl"] != "target":
                    continue
                if self.heading is None:
                    if line["msg"].startswith("*"):
                        self.heading = ""
                        self.description += line["msg"] + "\n"
                    else:
                        self.heading = line["msg"]
                else:
                    if len(line["msg"]) > 0:
                        self.description += line["msg"] + "\n"

    @property
    def name(self):
        name = self.heading
        if name is None or len(name) == 0:
            name = self.test_case_id
        return name

    @property
    def machine(self):
        return self.parent.machine

    @property
    def device(self):
        return self.parent.device

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def __str__(self):
        return self.name + ": " + str(self.result)

    def _search_log_for_test(self, log, log_end):
        """Loops reversely over the log and tries to find the relevant lines
        Returns the found line numbers.
        """

        # Search end of log output
        for line_no in range(log_end, len(log)):
            # logging.debug(line_no)

            line = log[line_no]
            if line["lvl"] == "results" and line["msg"]["case"] == self.test_case_id:
                log_end = line_no
                break

        if line_no >= len(log) - 1:
            raise LavaException("Failed to find test in log, skipping this test")

        log_start = None
        description_start = None
        description_end = None

        for line_no in range(log_end - 1, 0, -1):
            # logging.debug(line_no)
            line = log[line_no]

            # There seem to be two cases
            # Here the LAVA_SIGNAL_STARTRUN is used as start marker
            if line["lvl"] == "target" and line["msg"].startswith(
                "<LAVA_SIGNAL_STARTRUN"
            ):
                log_start = line_no
                break

            # Here the prior 'results' output is used as start marker
            # This is relevant when there are multiple tests in one
            # file
            if line["lvl"] == "results":
                log_start = line_no + 1

            # Search for the lava test description
            if line["lvl"] == "target" and line["msg"].startswith(
                "<TESTCASE_DESCRIPTION_END {}>".format(self.test_case_id)
            ):
                description_end = line_no

            if line["lvl"] == "target" and line["msg"].startswith(
                "<TESTCASE_DESCRIPTION_START {}>".format(self.test_case_id)
            ):
                description_start = line_no

            if (
                description_start is not None
                and description_end is not None
                and log_start is not None
            ):
                break
            if log_start is not None and line_no < log_start - 100:
                # Don't continue searching forever
                break

        # It seems the output of subsequent tests overlaps sometimes
        # The description may already start but the result of the prior
        # test comes few lines later
        if description_start is not None == log_start > description_start:
            log_start = description_start

        if log_start is None:
            raise LavaException("Failed to find test in log, skipping this test")

        return log_start, log_end, description_start, description_end


class LavaDevice:
    """Abstraction of a lava device"""

    def __init__(self, data, tags=None):
        self.name = data["hostname"]
        tag_ids = data["tags"]
        self.data = data
        self.tags = []
        self._serial = None
        self._article_no = None
        self._hw_revision = None

        if tags is not None:
            for t in tag_ids:
                if t in tags.keys():
                    tag = tags[t]
                    self.tags.append(tag)
                    if tag.is_serial:
                        self._serial = tag
                    elif tag.is_article_no:
                        self._article_no = tag
                    elif tag.is_hw_revision:
                        self._hw_revision = tag

    @property
    def machine(self):
        return re.sub(r"\d", "", self.name)

    @property
    def serial(self):
        if self._serial is None:
            return ""
        return self._serial.name

    @property
    def hw_revision(self):
        if self._hw_revision is None:
            return ""
        return self._hw_revision.name

    @property
    def article_no(self):
        if self._article_no is None:
            return ""
        return self._article_no.name

    def __lt__(self, other):
        return self.name < other.name

    @staticmethod
    def devicelist_from_lava(lava, tags=None, filter_types=None):
        if filter_types is None:
            filter_types = ["karl", "lxc"]
        devices = lava.get_device_list()
        device_list = {}
        for d in devices["results"]:
            if d["device_type"] in filter_types:
                continue
            device_list[d["hostname"]] = LavaDevice(d, tags)
        return device_list

    def __str__(self):
        return "LavaDevice " + self.name


class LavaTag:
    """Abstraction of a lava device"""

    def __init__(self, data):
        self.name = data["name"]
        self.description = data["description"]
        self.id = data["id"]

    def __str__(self):
        return "LavaTag " + self.name

    @property
    def is_serial(self):
        return re.match(r"^\d{7,10}$", self.name) is not None

    @property
    def is_article_no(self):
        return re.match(r"^900-\d{4}R\w?$", self.name) is not None

    @property
    def is_hw_revision(self):
        return re.match(r"^v\d{1,2}\.\d{1,3}(\.\d{1,3})?$", self.name) is not None

    @staticmethod
    def taglist_from_lava(lava):
        tags = lava.get_tag_list()
        tag_list = {}
        for t in tags["results"]:
            tag_list[t["id"]] = LavaTag(t)
        return tag_list


class LavaSuite:
    """Abstraction of lava suite"""

    def __init__(self, parent, suite, joblog):
        self._suite = suite
        self.lava = parent.lava
        self.parent = parent
        self.tests = []
        self.name = self._suite["name"]
        s = self.name.split("_", 2)
        if len(s) > 1:
            self.name = s[1]

        for test in self._get_tests():
            try:
                lava_test = LavaTest(self, test, joblog)
                self.tests.append(lava_test)
            except LavaException as e:
                logging.debug("%s: %s", str(e), test["name"])

        self.tests.sort()

    @property
    def machine(self):
        return self.parent.machine

    @property
    def device(self):
        return self.parent.device

    def _get_tests(self):
        ok, text = self.lava.plain_request(self._suite["resource_uri"] + "tests/")
        if not ok:
            raise LavaException("Failed to query lava for suite's tests")

        return json.loads(text)["results"]

    def __lt__(self, other):
        return self.tests[0] < other.tests[0]

    def __str__(self):
        return "TestSuite " + self.name


class LavaJob:
    """Abstraction for a lava job containing multiple tests
    on one device.
    """

    def __init__(self, lava, jobid, device_list=None):
        """Queries all needed information using the given
        lava class and stores
        """
        self.jobid = jobid
        self.lava = lava

        self._jobdetails = self.lava.get_job_details(jobid)

        self._suites = self.lava.get_job_suites(jobid)["results"]
        self._test_log = self.lava.get_test_log(jobid)

        self.name = self._jobdetails["description"]
        self.device = self._jobdetails["actual_device"]
        # Remove the number from device name to get the machine
        self.machine = re.sub(r"\d", "", self.device)
        self.device_type = self._jobdetails["requested_device_type"]

        if device_list is not None:
            if self.device in device_list:
                self.device = device_list[self.device]

        self.ok = (
            self._jobdetails["state"] == "Finished"
            and self._jobdetails["health"] == "Complete"
        )
        self.suites = []

        if not self.ok:
            logging.warning(
                "The job %d is not finished or it infrastructure problems", jobid
            )
            return

        for suite in self._suites:
            # Skip the internal lava steps
            if suite["name"] == "lava":
                continue
            try:
                lava_suite = LavaSuite(self, suite, self._test_log)
                self.suites.append(lava_suite)
            except LavaException as e:
                logging.debug("%s: %s", str(e), suite["name"])

        self.suites.sort()

    def __str__(self):
        return self.name


def main():
    host = get_lava_host()
    _, token = get_lava_credentials(host)
    lava = Lava(host, token)
    LavaJob(lava, "25142")


if __name__ == "__main__":
    main()
