#!/usr/bin/env python3
# Queries the test results from lava and generates a report
#

import argparse
import logging

from markdown_generator import MarkDownDoc, MarkDownTable
from lava_api import Lava, LavaTag, LavaDevice, LavaJob
from lava_credentials import get_lava_host, get_lava_credentials


class TestCollection:
    """Collection of tests with the same id, but different suites ..."""

    def __init__(self, test_case_id, tests):
        self.test_case_id = test_case_id

        self.tests = {}  # Tests sorted by machine
        for t in tests:
            if t.machine not in self.tests.keys():
                self.tests[t.machine] = []
            self.tests[t.machine].append(t)

    @property
    def has_measurements(self):
        for key in self.tests:
            for t in self.tests[key]:
                m = t.measurement
                if m is not None and len(m) > 0:
                    return True
        return False

    def get_measurements(self, machine):
        if machine not in self.tests.keys():
            return ""
        tests = self.tests[machine]
        out = []
        for t in tests:
            out.append("{:.2f}".format(float(t.measurement)))
        if len(out) == 0:
            return None
        return "/".join(out)

    @property
    def name(self):
        t = self.get_test()
        if t is None:
            return self.test_case_id
        return t.name

    # Sort alphabetically
    def __lt__(self, other):
        return str(self.get_test()) < str(other.get_test())

    def get_test(self, index=0):
        keys = list(self.tests.keys())
        if len(keys) < 1:
            return None
        if len(self.tests[keys[index]]) < 1:
            return None
        return self.tests[keys[index]][0]

    def get_result(self, machine):
        if machine not in self.tests.keys():
            return ""
        tests = self.tests[machine]
        count = 0
        for t in tests:
            if t.result:
                count += 1
        if count == len(tests):
            return "Passed"
        if count == 0:
            return "Failed"
        return "Partly failed ({} of {})".format(len(tests) - count, len(tests))

    @staticmethod
    def collections_from_suites(suite_list):

        all_tests = []
        for suite in suite_list:
            all_tests += suite.tests

        tests = {}
        test_collections = []
        for test in all_tests:
            if test.test_case_id not in tests.keys():
                tests[test.test_case_id] = []
            tests[test.test_case_id].append(test)
        for k, v in tests.items():
            test_collections.append(TestCollection(k, v))
        return test_collections

    def __str__(self):
        return "TestCollection " + self.name


class SuiteCollection:
    """Collection of lava suites like a row in a table, all suites have the same name"""

    def __init__(self, name, suites):
        self.name = name

        self.suites = {}  # Test suites sorted by machine
        for s in suites:
            if s.machine not in self.suites.keys():
                self.suites[s.machine] = []
            self.suites[s.machine].append(s)
        self.test_collections = TestCollection.collections_from_suites(suites)
        self.test_collections.sort()

    @staticmethod
    def markdown_result_table(suite_collections):
        # Write a header for a table
        all_machines = SuiteCollection.all_machines(suite_collections)

        table = MarkDownTable()
        table.set_alignment("llc")
        header = table.get_header()

        header.add_content(["Testsuite", "Test"], bold=True)
        for m in all_machines:
            header.add_content(m.capitalize(), bold=True)

        # Result rows
        for c in suite_collections:
            suite_col = c.name
            for t in c.test_collections:
                row = table.get_row()
                row.add_content([suite_col, "[{}](#{})".format(t.name, t.test_case_id)])
                for m in all_machines:
                    row.add_content(t.get_result(m))
                suite_col = ""

        return table

    @property
    def machines(self):
        return self.suites.keys()

    @property
    def devices(self):
        devices = []
        for l in self.suites.values():
            for s in l:
                d = s.device
                if d not in devices:
                    devices.append(d)
        return devices

    # Sort alphabetically
    def __lt__(self, other):
        return self.name < other.name

    def __str__(self):
        return "SuiteCollection " + self.name

    @staticmethod
    def collections_from_jobs(job_list):

        all_suites = []
        for job in job_list:
            if not job.ok:
                continue
            all_suites += job.suites
        suites = {}
        suite_collections = []
        for suite in all_suites:
            if suite.name not in suites.keys():
                suites[suite.name] = []
            suites[suite.name].append(suite)
        for k, v in suites.items():
            suite_collections.append(SuiteCollection(k, v))

        suite_collections.sort()
        return suite_collections

    @staticmethod
    def all_machines(collections):
        """Return a list of all machines (santino/santoka/...) used in the given collections"""
        machines = []
        for s in collections:
            for m in s.machines:
                if m not in machines:
                    machines.append(m)
        machines.sort()
        return machines

    @staticmethod
    def all_devices(collections):
        """Return a list of all devices (santino01/santino02/santoka03/...) used in the given collections"""
        devices = []
        for s in collections:
            for m in s.devices:
                if m not in devices:
                    devices.append(m)
        devices.sort()
        return devices


def lava_create_testreport(jobids, lava=None):
    # Get credentials
    if lava is None:
        host = get_lava_host()
        _, token = get_lava_credentials(host)
        lava = Lava(host, token)

    lava_tags = LavaTag.taglist_from_lava(lava)
    lava_devices = LavaDevice.devicelist_from_lava(lava, lava_tags)

    lava_jobs = []
    for jobid in jobids:
        lava_jobs.append(LavaJob(lava, jobid, lava_devices))

    suite_collections = SuiteCollection.collections_from_jobs(lava_jobs)

    all_machines = SuiteCollection.all_machines(suite_collections)
    all_devices = SuiteCollection.all_devices(suite_collections)

    # Write out the markdown
    doc = MarkDownDoc()
    doc.add_header("Test report")
    doc.add_text_block(
        "These tests are automatically executed on the given devices using lava as test engine."
    )

    doc.add_header("Overview table with test results", level=1)
    table = SuiteCollection.markdown_result_table(suite_collections)
    doc.add_element(table)

    for j in lava_jobs:
        if not j.ok:
            doc.add_text_block(
                "NOTE: Lava job {} for device {} did not complete, results are not part of the table".format(
                    j.jobid, j.machine.capitalize()
                )
            )

    doc.add_header("Test details")
    for c in suite_collections:
        doc.add_header("Test suite {}\n".format(c.name), level=2)
        for t in c.test_collections:
            doc.add_anchor(t.test_case_id)
            collapse = doc.add_collapse_block(t.name)
            collapse.add_text_block(t.get_test().description)
            if t.has_measurements:
                collapse.add_text_block("Measurements", italic=True)
                table = collapse.add_table()
                header = table.get_header()
                row = table.get_row()

                for m in all_machines:
                    header.add_content(m.capitalize(), bold=True)
                    row.add_content(t.get_measurements(m))
            collapse.add_text_block("_Devices_")
            collapse.add_text_block("Test was executed on following devices:")
            passed = []
            failed = []
            for d in all_devices:
                if d.machine in t.tests.keys():
                    for test in t.tests[d.machine]:
                        if test.result:
                            passed += [d.name.capitalize()]
                        else:
                            failed += [d.name.capitalize()]

            if len(passed):
                collapse.add_text_block("**Passed:**\n" + ", ".join(passed))
            if len(failed):
                collapse.add_text_block("**Failed:**\n" + ", ".join(failed))

    doc.add_header("Test devices")
    doc.add_text_block("These devices where used during the test run.")

    cols = 0
    for d in all_devices:
        if cols == 0:
            table = doc.add_table()
            header = table.get_row()
            row_article_no = table.get_row()
            row_serial = table.get_row()
            row_hw_revision = table.get_row()
            header.add_content("")
            row_serial.add_content("Serial Number", bold=True)
            row_hw_revision.add_content("Hardware Revision", bold=True)
            row_article_no.add_content("Article Number", bold=True)

        header.add_content(d.name.capitalize(), bold=True)
        row_serial.add_content(d.serial)
        row_hw_revision.add_content(d.hw_revision)
        row_article_no.add_content(d.article_no)
        cols += 1
        if cols > 5:
            cols = 0

    return doc.to_string()


def main():
    """
    Create a markdown report for the given lava jobids
    """

    # Read job file from command line
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "jobids", type=int, nargs="+", help="List of testjob IDs to query"
    )
    parser.add_argument(
        "--verbose",
        help="""More verbose output.""",
        action="store_true",
        default=False,
    )
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=10)
    else:
        logging.basicConfig(level=40)

    print(lava_create_testreport(args.jobids))


if __name__ == "__main__":
    main()
