#!/usr/bin/env python3
""" Function to read lava creadentials from environment, configfile or query the user on stdin """

import os
import configparser
import getpass
import sys
import stat

DEFAULT_LAVA_HOST = "lava"
DEFAULT_LAVA_CREDENTIAL_FILE = "~/.lava_credentials"


def get_lava_credentials(host=None):
    """
    Look for credentials in the following order:
        1. environment variables
        2. configuration file
        3. prompt the user
    If new data has been entered, save it to configuration file
    """

    if host is None:
        host = get_lava_host()

    configfile = os.path.expanduser(
        os.getenv("LAVA_CREDENTIAL_FILE", DEFAULT_LAVA_CREDENTIAL_FILE)
    )

    config = configparser.ConfigParser()
    config.read(configfile)
    new_data = False
    if "LAVA_USER" in os.environ:
        lava_user = os.environ["LAVA_USER"]
    else:
        try:
            lava_user = config.get(host, "LAVA_USER")
        except configparser.Error:
            print(
                "Missing user name for '%s'. Please either set it via the LAVA_USER"
                "environment variable or enter it now." % host
            )
            lava_user = input("LAVA_USER: ")
            new_data = True
    if "LAVA_TOKEN" in os.environ:
        lava_token = os.environ["LAVA_TOKEN"]
    else:
        try:
            lava_token = config.get(host, "LAVA_TOKEN")
        except configparser.Error:
            print(
                "Missing authentication token for '%s@%s'. Please either set it via the"
                "LAVA_TOKEN environment variable or enter it now." % (lava_user, host)
            )
            lava_token = getpass.getpass("LAVA_TOKEN: ")
            new_data = True

    # Ask the user whether to save credentials in private config file
    if new_data:
        valid = {"yes": True, "ye": True, "y": True, "": True, "no": False, "n": False}
        choice = None
        while not choice in valid:
            sys.stdout.write(
                "Do you want to save these credentials in '%s'? " "[Y/n] " % configfile
            )
            choice = input().lower()
        if valid[choice]:
            if not config.has_section(host):
                config.add_section(host)
            config.set(host, "LAVA_USER", lava_user)
            config.set(host, "LAVA_TOKEN", lava_token)
            os.chmod(configfile, stat.S_IRUSR | stat.S_IWUSR)
            with open(configfile, "w", encoding="utf-8") as cfgfile:
                config.write(cfgfile)

    return lava_user, lava_token


def get_lava_host():
    """Get LAVA host from environment or use default value"""
    host = os.getenv("LAVA_HOST", DEFAULT_LAVA_HOST)
    return host


def main():
    """Just test the credential functions"""
    host = get_lava_host()
    # Get credentials
    get_lava_credentials(host)


if __name__ == "__main__":
    main()
